/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 * Copyright (c) 2004-2005 Timothy Wong
 */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "debug.h"
#include "hash.h"

#define FAIL_ON_DUPE 0
#define ALLOW_DUPE 1
#define REPLACE_DUPE 2

#define DEFAULT_HASH_SIZE 1024
#define MAX_HASH_SIZE 1000000

//For the hash function
#define PRIME 2147483647 // (2^31) - 1
#define CONSTANT 3459

#if MAX_HASH_SIZE > PRIME
#error MAX_HASH_SIZE CANNOT BE GREATER THAN PRIME
#endif

typedef struct hash_entry_s {
  void* key;
  int keylen;
  void* value;
  struct hash_entry_s* prev;
  struct hash_entry_s* next;
} hash_entry_t;

//Private Functions
static int _hash_function(hash_t* hash,void* key,int len);
static hash_entry_t* _hash_entry_find(hash_t* hash,void* key,int len);
static int _hash_add(hash_t* hash,void* key,int keylen,void* value,int flag);
static int _hash_rehash(hash_t* hash);

int hash_init(hash_t* hash,int size)
{
  int ret = 0;
  print_entry_location();

  assert(hash != NULL);

  if(size <= 0)
    size = DEFAULT_HASH_SIZE;
  hash->table = malloc(sizeof(hash_entry_t) * size);
  hash->table_tail = malloc(sizeof(void*) * size);

  //Zero out all hash->table addresses and tail references
  memset(hash->table,0,sizeof(hash_entry_t*)*size);
  memset(hash->table_tail,0,sizeof(void*)*size);

  hash->tablesize = size;
  hash->number_of_objects = 0;

  print_exit_location();
  return ret;
}

//Does not copy the value, only stores the address
static int _hash_add(hash_t* hash,void* key,int keylen,void* value,int flag)
{
  int ret = 0;
  int index;
  hash_entry_t* entry;
  hash_entry_t* hptr;
  print_entry_location();

  assert(hash != NULL);
  assert(key != NULL && keylen > 0);

  //If number_of_objects > (tablesize * .75) then rehash
  if(hash->number_of_objects > ((double)hash->tablesize *.75))
    _hash_rehash(hash);

  index = _hash_function(hash,key,keylen);

  //Find End of List
  hptr = hash->table[index];
  if(hptr == NULL) {
    //Empty List
    hptr = hash->table[index];
    goto fill;
  }

  if(flag != ALLOW_DUPE) { //if ALLOW_DUPE then don't check for dupes
    while(hptr->next != NULL) {
      if(keylen == hptr->keylen && !memcmp(hptr->key,key,keylen)) {
	switch(flag) {
	case FAIL_ON_DUPE:
	  //If duplicate do not insert
	  ret = -1;
	  goto exit;
	case REPLACE_DUPE:
	  //If duplicate, replace
	  //XXX tim: could leak memory if value actually points to a heap address
	  hptr->value = value;
	  goto exit;
	}
      }
      hptr = hptr->next;
    }
  }
  else
    hptr = hash->table_tail[index];

  //Fill Entry
 fill:
  entry = malloc(sizeof(hash_entry_t));
  if(entry == NULL) {
    ret = ENOMEM;
    print_error_nomem();
    goto exit;
  }

  memset(entry,0,sizeof(hash_entry_t));
  entry->key = malloc(keylen);
  if(entry->key == NULL) {
    ret = ENOMEM;
    print_error_nomem();
    goto exit;
  }
  memcpy(entry->key,key,keylen);
  entry->keylen = keylen;
  entry->value = value;

  //Insert Entry
  if(hptr != NULL) {
      hptr->next = entry;
      entry->prev = hptr;
  }
  else
    hash->table[index] = entry;

  hash->table_tail[index] = entry;

 exit:
  if(!ret)
    hash->number_of_objects++;
  print_exit_location();
  return ret;
}

inline int hash_add(hash_t* hash,void* key,int keylen,void* value)
{
  return _hash_add(hash,key,keylen,value,FAIL_ON_DUPE);
}

inline int hash_add_allow_dupes(hash_t* hash,void* key,int keylen,void* value)
{
  return _hash_add(hash,key,keylen,value,ALLOW_DUPE);
}

inline int hash_add_replace_dupes(hash_t* hash,void* key,int keylen,void* value)
{
  return _hash_add(hash,key,keylen,value,REPLACE_DUPE);
}

void* hash_remove(hash_t* hash,void* key,int keylen)
{
  void* ret = NULL;
  hash_entry_t* hptr = NULL;
  int index = _hash_function(hash,key,keylen);
  print_entry_location();

  assert(hash != NULL);
  assert(key != NULL && keylen > 0);

  hptr = _hash_entry_find(hash,key,keylen);
  if(hptr != NULL) {
    if(hptr->prev != NULL) //Remove from linked-list
      hptr->prev->next = hptr->next;
    else {
      //First Element
      if(hptr->next == NULL)
	hash->table[index] = NULL;
      else
	hash->table[index] = hptr->next;
    }
    if(hptr->next != NULL)
      hptr->next->prev = hptr->prev;

    //Fix Tail Reference
    if(hash->table_tail[index] == hptr)
      hash->table_tail[index] = hptr->prev;

    ret = hptr->value;
    free(hptr->key);
    free(hptr);
  }

  if(ret != NULL)
    hash->number_of_objects--;

  print_exit_location();
  return ret;
}

void* hash_find(hash_t* hash,void* key,int keylen)
{
  void* ret = NULL;
  hash_entry_t* hptr = _hash_entry_find(hash,key,keylen);
  print_entry_location();

  assert(hash != NULL);
  assert(key != NULL && keylen > 0);

  if(hptr != NULL)
    ret = hptr->value;

  print_exit_location();
  return ret;
}

int hash_clear(hash_t* hash)
{
  hash_entry_t* hptr = NULL;
  hash_entry_t* prev = NULL;
  int index = 0;
  print_entry_location();

  assert(hash != NULL);

  for(index = 0;index < hash->tablesize;index++) {
    hptr = hash->table[index];
    while(hptr != NULL) {
      prev = hptr;
      hptr = hptr->next;
      free(prev->key);
      free(prev);
    }
    hash->table[index] = NULL;
    hash->table_tail[index] = NULL;
  }
  hash->number_of_objects = 0;

  print_exit_location();
  return 0;
}

int hash_destroy(hash_t* hash)
{
  print_entry_location();

  assert(hash != NULL);

  hash_clear(hash);
  free(hash->table);
  hash->table = NULL;

  free(hash->table_tail);
  hash->table_tail = NULL;

  print_exit_location();
  return 0;
}

//Destroys the hash and frees the hashed objects
int hash_destroy_all(hash_t* hash)
{
  hash_entry_t* hptr = NULL;
  hash_entry_t* prev = NULL;
  int index = 0;
  print_entry_location();

  assert(hash != NULL);

  for(index = 0;index < hash->tablesize;index++) {
    hptr = hash->table[index];
    while(hptr != NULL) {
      prev = hptr;
      hptr = hptr->next;
      if(prev->value != NULL)
	free(prev->value);
      free(prev->key);
      free(prev);
    }
  }

  free(hash->table);
  hash->table = NULL;

  free(hash->table_tail);
  hash->table_tail = NULL;
  hash->number_of_objects = 0;

  print_exit_location();
  return 0;
}

static int _hash_function(hash_t* hash,void* key,int len)
{
  int ret;
  unsigned long sum = 0;
  char* key_str = key;
  int i;
  print_entry_location();

  assert(key != NULL && len > 0);

  for(i = 0;i < len;i++)
    sum += key_str[i];

  if(sum == 0)
    ret = 0;

  sum = (sum * CONSTANT) % PRIME;
  ret = sum % hash->tablesize;

  print_exit_location();
  return ret;
}

static hash_entry_t* _hash_entry_find(hash_t* hash,void* key,int len)
{
  hash_entry_t* ret = NULL;
  int index = _hash_function(hash,key,len);
  print_entry_location();

  assert(hash != NULL);

  for(ret = hash->table[index];ret != NULL;ret = ret->next) {
    if(ret->keylen != len)
      continue;
    if(!memcmp(ret->key,key,len))
      goto exit;
  }

 exit:
  print_exit_location();
  return ret;
}

static int _hash_rehash(hash_t* hash)
{
  int err = 0;
  int index;
  int original_number_of_objects;
  int original_table_size;
  hash_entry_t** original_table;
  hash_entry_t* hptr = NULL;
  hash_entry_t* prev = NULL;

  print_entry_location();
  assert(hash != NULL);

  if(hash->tablesize * 2 > MAX_HASH_SIZE)
    goto out;

  original_number_of_objects = hash->number_of_objects;
  original_table = hash->table;
  original_table_size = hash->tablesize;

  //Double the size of the table
  hash_init(hash,original_table_size*2);

  //Rehash
  for(index = 0;index < original_table_size;index++) {
    hptr = original_table[index];
    while(hptr != NULL) {
      err = hash_add_allow_dupes(hash,hptr->key,hptr->keylen,hptr->value);
      assert(!err);
      prev = hptr;
      hptr = hptr->next;
      free(prev);
    }
  }
  free(original_table);
  assert(original_number_of_objects == hash->number_of_objects);

 out:
  print_exit_location();
  return err;
}
