/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 */

#include "idriver.h"
#include <openssl/evp.h>
#include "common.h"
#include "od_encrypt.h"
#include <string.h>

static char iv[EVP_MAX_IV_LENGTH];
static EVP_CIPHER_CTX ctx;
static char *ct_buf = NULL;
static int   ct_buf_size;
static char *et_buf = NULL;
static int   et_buf_size;
static int  ended = 0;
static const EVP_CIPHER *type;
static char *ct_begin;
static char *ct_end;


static int cryptdriver_init(struct parser *p, void *params, struct idriver *next)
{
	struct decrypt_params *dp = (struct decrypt_params *)params;
	int ret = 0;
	EVP_CIPHER_CTX_init(&ctx);
	if (!next) {
		snprintf(p->err, ERR_SIZE, "decrypt driver requires a next driver");
		ret = ERR_INVAL;
		goto out;
	}
	if (!dp || dp->keylen > EVP_MAX_KEY_LENGTH) {
		snprintf(p->err, ERR_SIZE, "Improper key specified");
		ret = ERR_INVAL;
		goto out;
	}
	memset(iv, 0, sizeof(iv));
	type = EVP_bf_cbc();
/*
	if ((type = EVP_get_cipherbyname(SN_bf_cbc)) == NULL) {
		snprintf(p->err, ERR_SIZE, "cipher not found");
		ret = ERR_INVAL;
		goto out;
	}
*/
	if (!EVP_DecryptInit(&ctx, type, (unsigned char *)dp->key,(unsigned char *) iv)) {
		snprintf(p->err, ERR_SIZE, "decrypt_init failed");
		ret =  ERR_INVAL;
		goto out;
	}
	et_buf_size = 4096;
	ct_buf_size = et_buf_size + EVP_CIPHER_key_length(type);
	if ((ct_buf = malloc(ct_buf_size)) == NULL) {
		snprintf(p->err, ERR_SIZE, "Could not allocate compressed text buffer");
		ret = ERR_NOMEM;
		goto out;
	}
	if ((et_buf = malloc(et_buf_size)) == NULL) {
		snprintf(p->err, ERR_SIZE, "Could not allocate compressed text buffer");
		ret = ERR_NOMEM;
		goto out;
	}
	ct_begin = ct_end = ct_buf;
	ended = 0;
	decrypt_idriver->next_driver = next;
	return ret;
out:
	if (et_buf)
		free(et_buf);
	if (ct_buf)
		free(ct_buf);
	EVP_CIPHER_CTX_cleanup(&ctx);
	return ret;
}

static int cryptdriver_read(struct parser *p, char *buffer, int len)
{
	int read = 0;
	char *ptr = buffer;
	int in_read;
	int avail;
	int outl;

	while (len > 0) {
		// read from the cleartext buffer
		avail = ct_end - ct_begin;
		if  (avail > 0) {
			if (len > avail) {
				memcpy(ptr, ct_begin, avail);
				len -= (ended) ? len : avail;
				read += avail;
				ptr += avail;
				ct_begin = ct_end = ct_buf;
			} else {
				memcpy(ptr, ct_begin, len);
				ct_begin += len;
				read += len;
				ptr += len;
				len = 0;
			}
		} else if (avail == 0) {
			if (ended)
				len = 0;
			ct_begin = ct_end = ct_buf;
		}
		if (len && !ended) {
			in_read = decrypt_idriver->next_driver->ops->read(p, et_buf, et_buf_size);
			if (in_read < 0) {
				snprintf(p->err, ERR_SIZE, "Error while reading: %d", errno);
				return -ERR_IO;
			}
			if (in_read != et_buf_size)
				ended = 1;
			if (in_read) {
				if (!EVP_DecryptUpdate(&ctx, (unsigned char *)ct_buf, &outl,(unsigned char *) et_buf, in_read)) {
					snprintf(p->err, ERR_SIZE, "Error while decrypting");
					return -ERR_IO;
				}
				ct_end = ct_begin + outl;
			}
			if (ended) {
				if (!EVP_DecryptFinal(&ctx,(unsigned char*) ct_end, &outl)) {
					snprintf(p->err, ERR_SIZE, "Error while completing decryption");
					return -ERR_IO;
				} else {
					ct_end += outl;
				}
			}
		}
	}
	return read;


}

static void cryptdriver_release()
{
	free(ct_buf);
	free(et_buf);
	EVP_CIPHER_CTX_cleanup(&ctx);

}

struct input_ops cryptdriver_ops  = {
	init: cryptdriver_init,
	read: cryptdriver_read,
	release: cryptdriver_release
};

struct idriver cryptdriver = {
	ops : &cryptdriver_ops
};

struct idriver *decrypt_idriver = &cryptdriver;
