/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_compress.c
 * output driver for compression
 */

#include "tracer.h"
#include "od_compress.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include "zutil.h"
#else
#include "zlib/zutil.h"
#endif
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/*
 * zcalloc implementation for zlib
 * simply calls KMALLOC with GFP_KERNEL
 */
void *zcalloc (void *opaque, unsigned num, unsigned size)
{
        return (void *)KMALLOC(num * size, GFP_KERNEL);
}
/*
 * zcfree implementation for zlib
 * simply calls KFREE
 */
void zcfree (void *opaque, void *ptr)
{
        KFREE(ptr);
}

/*
 * internal structure for this driver
 */

struct od_compress_data {
	char *buf;
	int buf_size;
	char *in_buf;
	int   in_buf_size;
	int   in_avail;
	z_stream stream;

};

/**
 * Destroys the checompress driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_compress_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the file driver object.
 * @param Checompress driver object to initialize
 * @param <code>struct od_compress_params</code> type initialization parameters
 * @see struct od_compress_params
 */

static int compress_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = 0;
	struct od_compress_data *cdata;
	z_stream *stream;
	struct od_compress_params *cp = (struct od_compress_params *)params;
	ASSERT(this->od_private_data == NULL);
	print_entry_location();
	if (!next || !cp) {
		ret = -EINVAL;
		goto out;
	}
	ret = -ENOMEM;

	cdata = KMALLOC(sizeof(struct od_compress_data), GFP_KERNEL);
	if (!cdata) {
		goto out;
	}
	memset(cdata, 0, sizeof(*cdata));

	cdata->buf_size = next->od_ops->getblocksize(next);
	if ((cdata->buf = vmalloc(cdata->buf_size)) == NULL) {
		goto out_free;
	}
	/* If we are not at the top (level != 0), we need not do input buffering */
	if (level == 0 && cp->in_buf_size) {
		cdata->in_buf_size = cp->in_buf_size;
		if ((cdata->in_buf = vmalloc(cdata->in_buf_size)) == NULL) {
			goto out_free;
		}
	}
	cdata->in_avail = cdata->in_buf_size;
	stream = &cdata->stream;
	stream->zalloc = (alloc_func)0;
	stream->zfree = (free_func)0;
	stream->opaque = (voidpf)0;

	if (deflateInit(stream, Z_DEFAULT_COMPRESSION) != Z_OK) {
		goto out_free;
	}
	ret = 0;
	stream->next_out = cdata->buf;
	stream->avail_out = cdata->buf_size;
	stream->avail_in = 0;
	stream->next_in = NULL;

	this->od_private_data = cdata;

	this->od_next = next;
	goto out;

out_free:
	if (cdata->in_buf)
		vfree(cdata->in_buf);
	if (cdata->buf)
		vfree(cdata->buf);
	KFREE(cdata);
out:
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void compress_release(struct output_driver *this)
{
	struct od_compress_data *cdata = (struct od_compress_data *)this->od_private_data;
	ASSERT(this);
	ASSERT(this->od_type == OD_COMPRESS);
	if (deflateEnd(&cdata->stream) != Z_OK) {
		fist_dprint(2, "Error while releasing compression stream");
	}
	if (cdata->in_buf)
		vfree(cdata->in_buf);
	vfree(cdata->buf);
	KFREE(cdata);
	this->od_private_data = NULL;
}

/**
 * flushes the internal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int compress_flush(struct output_driver *this)
{
	struct od_compress_data *cdata = (struct od_compress_data *)this->od_private_data;
	z_stream *stream = &cdata->stream;
	int ret = 0;
	int done = 0;
	int z_ret = Z_OK;
	int total = 0;
	int to_write;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_COMPRESS);
	ASSERT(this->od_next);

	if (cdata->in_buf && ((to_write = cdata->in_buf_size - cdata->in_avail) > 0)) {
		stream->avail_in = to_write;
		stream->next_in = cdata->in_buf;
		while (stream->avail_in) {
			if (stream->avail_out == 0) {
				stream->next_out = cdata->buf;
				if ((ret = this->od_next->od_ops->write(this->od_next, cdata->buf, cdata->buf_size,OO_KERNEL_BUFFER)) != cdata->buf_size) {
					break;
				}
				stream->avail_out = cdata->buf_size;
			}
			if (deflate((stream), Z_NO_FLUSH) != Z_OK) {
				ret = -EIO;
				break;
			}
		}
		if (ret < 0)
			goto out;
	}

	stream->avail_in = 0;
	for (;;) {
		to_write = cdata->buf_size - stream->avail_out;

		if (to_write != 0) {
			if ((ret = this->od_next->od_ops->write(this->od_next, cdata->buf, to_write,OO_KERNEL_BUFFER)) != to_write) {
				break;
			}
			total += ret;
			stream->next_out = cdata->buf;
			stream->avail_out = cdata->buf_size;
		}
		if (done) break;
		z_ret = deflate(stream, Z_FINISH);
		if (to_write == 0 && z_ret == Z_BUF_ERROR) z_ret = Z_OK;
		done = (stream->avail_out != 0 || z_ret == Z_STREAM_END);

		if (z_ret != Z_OK && z_ret != Z_STREAM_END) {
			ret = -EIO;
			break;
		}
	}

        if (ret >= 0)
                ret = this->od_next->od_ops->flush(this->od_next);
out:
	print_exit_status(ret);
	return ret;
}

/**
 * Writes the data to output.
 * If internal buffering is enabled, data is written to buffer
 * till the buffer runs out.
 * @param object to operate on
 * @param buffer to write out
 * @param number of bytes to write
 * @param whether the buffer is a user-level buffer (OO_USER_BUFFER)
 * @returns number of bytes written
 */

int compress_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
	int ret = 0;
	int to_write = len;
	int junk;
	struct od_compress_data *cdata = (struct od_compress_data *)this->od_private_data;
	z_stream *stream = &cdata->stream;
	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_COMPRESS);
	/* compression driver cannot work in isolation */
	ASSERT(this->od_next);

	stream->next_in = (Bytef *)buf;
	stream->avail_in = len;

	while (to_write > 0) {
		if (cdata->in_buf) {
			if (cdata->in_avail > 0) {
				int copy = (to_write < cdata->in_avail) ? to_write : cdata->in_avail;
				if(is_user_buffer == OO_KERNEL_BUFFER)
				  memcpy(cdata->in_buf + cdata->in_buf_size - cdata->in_avail, buf, copy);
				else
				  junk = copy_from_user(cdata->in_buf + cdata->in_buf_size - cdata->in_avail, buf, copy);
				buf += copy;
				to_write -= copy;
				cdata->in_avail -= copy;
				if (to_write == 0)
					break;
			}
			/* When we reach here, we've filled up the input buffer */
			stream->next_in = cdata->in_buf;
			stream->avail_in = cdata->in_buf_size;
		}
		if (stream->avail_out == 0) {
			stream->next_out = cdata->buf;
			if ((ret = this->od_next->od_ops->write(this->od_next, cdata->buf, cdata->buf_size,OO_KERNEL_BUFFER)) != cdata->buf_size) {
				break;
			}
			stream->avail_out = cdata->buf_size;
		}
		if (deflate((stream), Z_NO_FLUSH) != Z_OK) {
			ret = -EIO;
			break;
		}
		if (cdata->in_buf) {
		       if (stream->avail_in == 0)
				cdata->in_avail = cdata->in_buf_size;
		} else {
			to_write = stream->avail_in;
		}
	}

/*

	while (stream->avail_in != 0) {
		if (stream->avail_out == 0) {
			stream->next_out = cdata->buf;
			if ((ret = this->od_next->od_ops->write(this->od_next, cdata->buf, cdata->buf_size)) != cdata->buf_size) {
				break;
			}
			stream->avail_out = cdata->buf_size;
		}
		if (deflate((stream), Z_NO_FLUSH) != Z_OK) {
			ret = -EIO;
			break;
		}
	}
*/
//out:
	ret = (ret < 0) ? ret : len;
	print_exit_status(ret);
	return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
int compress_getblocksize(struct output_driver *this)
{
	ASSERT(this);
	ASSERT(this->od_type == OD_COMPRESS);
	return (this->od_next->od_ops->getblocksize(this->od_next));
}

static struct output_ops compress_od_ops = {
write: compress_write,
flush: compress_flush,
getblocksize: compress_getblocksize,
init: compress_init,
release: compress_release
};

/**
 * Creates a compression driver object.
 * @return A compression driver object with function pointers initialised
 */

struct output_driver* create_compress_driver()
{
	struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

	if (driver) {
		driver->od_ops = &compress_od_ops;
		driver->od_type = OD_COMPRESS;
		driver->od_private_data = NULL;
		driver->od_next = NULL;
	}
	return driver;
}


